/* Copyright 1998 Peter Seebach.  All rights reserved.
 * This program may be freely distributed under the terms of the "BSD"
 * license.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "opts.h"
#include "version.h"

int debug = 0;

static void input(char *, size_t, FILE *);

static int
usage(int exitval) {
	fprintf(stderr, "usage: moo [-m max] [-n digits] [-h] [numbers ...]\n");
	fprintf(stderr, "\t-m: set maximum for each digit (1-9, default 9)\n");
	fprintf(stderr, "\t-n: set number of digits (1-10, default 4)\n");
	fprintf(stderr, "\t-h: print this message.\n");
	exit(exitval);
}

int
main(int argc, char **argv) {
	char *number;
	char *guess;
	char *tnumber;
	char *spanset;
	int guesses;
	int numlen = 4, maxnum = 9;
	int o, i;

	while ((o = optsopt(argc, argv, "hdm:n:V")) != -1) {
		char *t; /* for error-checking args */

		switch (o) {
		case 'd':
			debug = 1;
			break;
		case 'h':
			usage(0);
			break;
		case 'm':
			maxnum = strtol(optsarg, &t, 10);
			if (maxnum < 1 || maxnum > 9 || *t) {
				usage(EXIT_FAILURE);
			}
			break;
		case 'n':
			numlen = strtol(optsarg, &t, 10);
			if (numlen < 1 || numlen > 10 || *t) {
				usage(EXIT_FAILURE);
			}
			break;
		case 'V':
			version();
			break;
		}
	}

	number = malloc(numlen + 1);
	guess = malloc(numlen + 3);
	tnumber = malloc(numlen + 1);
	spanset = malloc(maxnum + 2);
	if (!number || !guess || !tnumber || !spanset) {
		fprintf(stderr, "I can't believe it.  You don't have enough memory to play \"moo\".\n");
		exit(EXIT_FAILURE);
	}

	for (i = 0; i <= maxnum; ++i)
		spanset[i] = i + '0';
	spanset[i] = '\0';

	srand48(time(NULL));
	printf("enter q to quit\n");
	for(;;) {
		/* inititalize */
		guesses = 0;
		if (argv[optsind]) {
			snprintf(number, numlen + 1, "%s", argv[optsind++]);
			if (strspn(number, spanset) != numlen) {
				fprintf(stderr, "sorry, but %s is not a %d-digit number using digits 0-%d.\n",
					number, numlen, maxnum);
				exit(EXIT_FAILURE);
			}
		} else {
			for (i = 0; i < numlen; ++i)
				number[i] = '0' + lrand48() % maxnum;
			number[i] = '\0';
		}
		if (debug)
			printf("number is %s\n", number);

		for(;;) {
			int bulls = 0, cows = 0;
			int i, j;

			printf("your guess? ");
			fflush(stdout);
			input(guess, numlen + 3, stdin);
			if (strspn(guess, spanset) != numlen) {
				if (strchr(guess, 'q') || strchr(guess, 'Q'))
					break;
				printf("please enter a %d-digit number using only digits 0-%d.\n",
					numlen, maxnum);
				continue;
			}
			++guesses;
			strcpy(tnumber, number);
			for (i = 0; i < numlen; ++i) {
				if (guess[i] == tnumber[i]) {
					++bulls;
					guess[i] = '\0';
					tnumber[i] = '\0';
				}
			}
			for (i = 0; i < numlen; ++i) {
				if (guess[i] == '\0')
					continue;
				for (j = 0; j < numlen; ++j) {
					if (j == i)
						continue;
					if (guess[i] == tnumber[j]) {
						++cows;
						guess[i] = '\0';
						tnumber[j] = '\0';
						break;
					}
				}
			}
			if (bulls == numlen) {
				printf("You got it in %d guess%s.\n",
					guesses,
					guesses == 1 ? "" : "es");
				break;
			} else if (bulls || cows) {
				printf("%d bull%s, %d cow%s\n",
					bulls, bulls == 1 ? "" : "s",
					cows, cows == 1 ? "" : "s");
			} else {
				printf("mu.\n");
			}
		}
		for(;;) {
			printf("Another game? ");
			fflush(stdout);
			input(guess, 3, stdin);
			if (strchr(guess, 'n') || strchr(guess, 'N'))
				return 0;
			if (strchr(guess, 'y') || strchr(guess, 'Y'))
				break;
			printf("Please answer y or n.\n");
		}
	}
	return 0;
}

static void
input(char *s, size_t n, FILE *fp) {
	/* 4.4BSD function, discards input and output.
	 * does not flush output.  Not portable.
	 */
#if 0
	fpurge(fp);
#endif
	*s = '\0';
	if (!fgets(s, n, fp))
		exit(0);
}
