/* Started Dec 5, 1994 at Xerox.
 * All wrongs reversed. */
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef __USLC__
extern int tgetent(char *, char *);
extern int tgetnum(char *);
#else
#include <curses.h>
/*#include <term.h>*/
#endif

#include "opts.h"
#include "version.h"

/* c - put into cols (default determined by input) */

int cols = 0;
int longest = 0;
int width;
int hmode = 0, nspaces = 2;
int cw = 0;
int vflag = 0, fflag = 0;

/*@null@*/
char *buffer = 0;
/*@null@*/
size_t *lns = 0;
int size = 0;
int count = 0;
int lcount = 0;
int lsize = 0;
char *us;
void usage(void);
int spew_output(void);
int slurp_input(int, FILE *);
void callversion(option *, opt_value *);
void csetnum(option *, opt_value *);

void
usage(void) {
	fputs("usage: c [-V] [-n #] [-w #] [-c #] [-h123456789]\n", stderr);
	fputs("\t-n #: use # spaces between cols (default 2)\n", stderr);
	fputs("\t-c #: use # cols (default best fit)\n", stderr);
	fputs("\t-w #: use # characters per line (default stty or 80)\n", stderr);
	fputs("\t-#  : use # cols\n", stderr);
	fputs("\t-h  : adjacent items are left->right, not up->down\n", stderr);
	fputs("\t-V  : print version info\n", stderr);
}

int
spew_output(void) {
	int n;
	int citem;
	int i, j;

	--lcount;
	if (cols == 0) {
		cw = longest + nspaces;

		if (cw < nspaces)
			cw = width + nspaces;

		cols = width / cw;
		cw -= nspaces;

		if (cols == 0) {
			cols = 1;
			cw = width;
		}

		while ((cols + 1) * cw + (nspaces * cols) < width)
			++cols;
	}

	n = lcount / cols;

	if (lcount == 0)
		return 0;

	if (n * cols < lcount)
		++n;

	for (i = 0; i < n; ++i) {
		for (j = 0; j < cols; ++j) {
			if (hmode)
				citem = i * cols + j;
			else
				citem = i + (j * n);

			if (citem >= lcount) {
				continue; /* we are done */
			}

			if (hmode ?
				((j == cols - 1) || (citem == lcount - 1)) :
				(citem >= lcount - n)) {
				printf("%s", buffer + lns[citem]);
			} else {
				printf("%-*s%*s", cw,
					buffer + lns[citem],
					nspaces, "");
			}
		}

		printf("\n");
	}
	free(lns);
	lns = 0;
	free(buffer);
	buffer = 0;
	lsize = 0;
	size = 0;
	return 0;
}

/* Slurp input.  Our goal is to always leave this function in a state such
 * that count refers (accurately) to the number of lns.  Lines should be
 * an allocated array of count size_t's, offsets into buffer, each `pointing'
 * to a null terminated string.
 * The slurping routine will eat tabs, turning them into multiples-of-8
 * spaces.  Obviously, we need many counters to do this:
 * 1.  Position in line.
 * 2.  Current line.
 * 3.  Position in buffer.
 * 4.  Size of buffer, and of lns[].
 *
 */
int
slurp_input(int llen, FILE *fp) {
	int lpos = 0, i, j, c;
	int formfeed = 0;

restart:
	if (!lns) {
		lns = malloc(256 * sizeof(size_t));
		if (!lns) {
			fprintf(stderr, "%s: couldn't allocate line memory.\n", us);
			exit(EXIT_FAILURE);
		}
		lcount = 0;
		lsize = 256;
	}
	if (!buffer) {
		buffer = malloc(4096);
		if (!buffer) {
			fprintf(stderr, "%s: couldn't allocate character memory.\n", us);
			exit(EXIT_FAILURE);
		}
		count = 0;
		size = 4096;
	}
	lns[lcount++] = 0;
	while ((c = getc(fp)) != EOF) {
		switch (c) {
		case '\f':
			formfeed = 1;
		case '\v':
		case '\r':
			/* ignore vertical tabs and carriage returns */
			break;
		case '\t':
			j = 8 - (lpos % 8);
			if (lpos + j > llen) {
				j = llen - lpos;
			}
			if (count + j > size) {
				buffer = realloc(buffer, size * 2);
				if (!buffer) {
					fprintf(stderr, "%s: couldn't reallocate character memory.\n", us);
					exit(EXIT_FAILURE);
				}
				size *= 2;
			}
			for (i = 0; i < j; ++i) {
				buffer[count++] = ' ';
			}
			lpos += j;
			c = ' ';
			/* FALLTHROUGH INTENTIONAL */
		default:
			buffer[count++] = (char) c;
			if (lpos++ < llen)
				break;
			
			/* if lpos >= llen, we now eat characters until EOF
			 * or \n, then pretend we just got a newline. */
			while ((c = getc(fp)) != EOF)
				if (c == '\n')
					break;
			--count;
			/*@fallthrough@*/
		case '\n':
			while ((lpos > 0) && (buffer[count - 1] == ' ')) {
				--lpos;
				--count;
			}
			buffer[count++] = '\0';
			if (lpos == 0) {
				if (formfeed && !fflag) {
					spew_output();
					putchar('\f');
					putchar('\n');
					formfeed = 0;
					longest = 0;
					cols = 0;
					goto restart;
				} else {
					count -= lpos;
					buffer[count++] = '\0';
					lpos = 1;
				}
			}
			formfeed = 0;
			longest = lpos > longest ? lpos : longest;
			lpos = 0;
			lns[lcount++] = count;
			if (lcount >= lsize) {
				lns = realloc(lns, lsize * 2 * sizeof(size_t));
				lsize *= 2;
				if (!lns) {
					fprintf(stderr, "%s: couldn't reallocate line memory.\n", us);
					exit(EXIT_FAILURE);
				}
			}
			break;
		}
		if (count >= size) {
			buffer = realloc(buffer, size * 2);
			if (!buffer) {
				fprintf(stderr, "%s: couldn't reallocate character memory.\n", us);
				exit(EXIT_FAILURE);
			}
			size *= 2;
		}
	}
	buffer[count] = '\0';
#ifdef DEBUG
	fprintf(stderr, "slurped %d/%d chars, %d/%d lines [longest %d]\n",
		count, size, lcount - 1, lsize, longest);
#endif
	return 0;
}

void
callversion(option *op, opt_value *ov) {
	version();
}

void
csetnum(option *op, opt_value *ov) {
	cols = op->abbr - '0';
}

option options[] = {
	{ OTYPE_INT, 'c', "cols", "number of columns",
		OFLAG_NONE, (opt_value *) &cols, 0, 0, 0 },
	{ OTYPE_BOL, 'f', "formfeed", "discard form feeds",
		OFLAG_NONE, (opt_value *) &fflag, 0, 0, 0 },
	{ OTYPE_BOL, 'h', "horizontal", "horizontal output",
		OFLAG_NONE, (opt_value *) &hmode, 0, 0, 0 },
	{ OTYPE_INT, 'n', "nspaces", "number of spaces",
		OFLAG_DFL, (opt_value *) &nspaces, 0, 0, 0 },
	{ OTYPE_INT, 'w', "width", "output width",
		OFLAG_NONE, (opt_value *) &width, 0, 0, 0 },
	{ OTYPE_NUL, '1', 0, "1 column",
		OFLAG_NONE, 0, 0, 0, csetnum },
	{ OTYPE_NUL, '2', 0, "2 cols",
		OFLAG_NONE, 0, 0, 0, csetnum },
	{ OTYPE_NUL, '3', 0, "3 cols",
		OFLAG_NONE, 0, 0, 0, csetnum },
	{ OTYPE_NUL, '4', 0, "4 cols",
		OFLAG_NONE, 0, 0, 0, csetnum },
	{ OTYPE_NUL, '5', 0, "5 cols",
		OFLAG_NONE, 0, 0, 0, csetnum },
	{ OTYPE_NUL, '6', 0, "6 cols",
		OFLAG_NONE, 0, 0, 0, csetnum },
	{ OTYPE_NUL, '7', 0, "7 cols",
		OFLAG_NONE, 0, 0, 0, csetnum },
	{ OTYPE_NUL, '8', 0, "8 cols",
		OFLAG_NONE, 0, 0, 0, csetnum },
	{ OTYPE_NUL, '9', 0, "9 cols",
		OFLAG_NONE, 0, 0, 0, csetnum },
	{ OTYPE_NUL, 'V', "version", "version information",
		OFLAG_NONE, 0, 0, 0, callversion },
	{ OTYPE_END, '\0', 0, 0,
		OFLAG_NONE, 0, 0, 0, 0 },
};

int
main(int argc, char *argv[])
{
	int i, retflag = 0;
	char *s;
	FILE *fp;

	s = argv[0];
	s += strlen(s) - 1;

	if (isdigit(*s))
		cols = *s - '0';
	
	s = argv[0];

	if ((us = strrchr(s, '/')))
		++us;
	else
		us = argv[0];

	{
		char *term;
		char tcp[2048];

		term = getenv("TERM");
		if (term) {
			if (tgetent(tcp, term) > 0) {
				width = tgetnum("co");
			}
		}
	}

	if (optsgets(argc, argv, options)) {
		usage();
		exit(EXIT_FAILURE);
	}

	if (width < 1)
		width = 80;

	if (vflag)
		version();

	if (cols != 0)
		cw = (width - nspaces * (cols - 1)) / cols;
	else
		cw = width;

	if (optsind < argc) {
		for (i = optsind; i < argc; ++i) {
			if (strcmp(argv[i], "-"))
				fp = fopen(argv[i], "r");
			else
				fp = stdin;
			if (!fp) {
				fprintf(stderr, "%s: error: couldn't open file '%s'.\n", us, argv[i]);
				retflag = 1;
			} else
				retflag |= slurp_input(cw, fp);
			if (fp && (fp != stdin))
				fclose(fp);
		}
	} else {
		retflag |= slurp_input(cw, stdin);
	}


	return retflag | spew_output();
}
